from pyezspark import gym_http_server
import gym
import ezclient
import threading
import re
import time
import requests
import json
import datetime

class localTunnelRun(threading.Thread):
    def __init__(self, ip, port):
        threading.Thread.__init__(self)
        self.ip = ip
        self.port = port
        self.url = localtunnel.get_localtunnel_url()
    
    def get_url(self):
        return self.url
        
    def run(self):
        localtunnel.main(self.ip, self.port, self.url)

class postTunnelRun(threading.Thread):
    def __init__(self, gym_server_url,training_public_key, training_private_key, neat, gym_game_name, buffer_size, genome_input, genome_output, url_name, remote_ip, remote_port, genomes_per_client):
        threading.Thread.__init__(self)
        self.gym_server_url = gym_server_url
        self.training_public_key = training_public_key
        self.training_private_key = training_private_key
        self.neat = neat
        self.gym_game_name = gym_game_name
        self.buffer_size = buffer_size
        self.input_size = genome_input
        self.output_size = genome_output
        self.url_name = url_name
        self.remote_ip = remote_ip
        self.remote_port = remote_port
        self.genomes_per_client = genomes_per_client
        self.client = ezclient.Client(self.training_public_key,training_private_key = self.training_private_key,neat_class = self.neat, gym_game_name = self.gym_game_name, buffer_size = self.buffer_size, genome_input = self.input_size, genome_output = self.output_size, url_name = self.url_name)
        self.headers = {'Content-type': 'application/json'}
    def run(self):
        self.client.connect(self.remote_ip,self.remote_port, genomes_per_client = self.genomes_per_client)
        while(True):
            # keep the communication active
            while(not self.client.is_disconnected()):
                self.client.host_direct_main_loop_http_requests()
                time.sleep(0.0001)
            # we have been disconnected by the server (bad requests or time out)
            if self.client.got_broken_pipe():
                self.client.connect(self.remote_ip, self.remote_port, genomes_per_client = self.genomes_per_client)
                continue
            # we are here, a trainer is communicating with us
            request_body = self.client.get_string().decode('utf-8')
            try:
                request_body = json.loads(request_body)
            except:
                self.client.set_body_http(0, 'null')
                continue
            if 'endpoint' not in request_body:
                self.client.set_body_http(0, 'null')
                continue
            endpoint = request_body['endpoint']
            request_body.pop('endpoint')

            if endpoint == '/v1/envs/':
                ret = gym_http_server.env_create(request_body)
            elif endpoint == '/v1/envs/step/':
                ret = gym_http_server.multi_step(request_body)
            else:
                self.client.set_body_http(0, 'null')
                continue
            self.client.set_body_http(1, ret)


class Host:
    def __init__(self, gym_game_name, alone_training_iterations, configuration_dict,
                 max_number_of_games, max_number_of_steps, training_public_key = None, training_private_key = None, socket1_training_public_key = None,
                 socket1_training_private_key = None,socket2_training_public_key = None,socket2_training_private_key = None,socket3_training_public_key = None,
                 socket3_training_private_key = None, t_val = 5, threads_for_alone_training = 4):
        if max_number_of_steps <= 0 or max_number_of_games <= 0 or alone_training_iterations < 0 or threads_for_alone_training < 1:
            print("Error, something among max number of steps, max number of games or alone training iterations is <= 0")
            exit(1)
        ezclient.get_randomness()
        self.t_val = t_val
        self.gym_game_name = gym_game_name
        self.input_size = configuration_dict['input_size']
        self.output_size = configuration_dict['output_size']
        self.threads_for_alone_training = threads_for_alone_training
        self.configuration_dict = configuration_dict
        if self.configuration_dict['minimum_reward'] == None:
            self.configuration_dict['minimum_reward'] = 0
        self.neat = ezclient.Neat(self.input_size, self.output_size, initial_population = configuration_dict['initial_population'],
                                  species_threshold = configuration_dict['species_threshold'],max_population = configuration_dict['max_population'],
                                  generations = configuration_dict['generations'], percentage_survivors_per_specie = configuration_dict['percentage_survivors_per_specie'], 
                                  connection_mutation_rate = configuration_dict['connection_mutation_rate'], new_connection_assignment_rate = configuration_dict['new_connection_assignment_rate'], 
                                  add_connection_big_specie_rate = configuration_dict['add_connection_big_specie_rate'], add_connection_small_specie_rate = configuration_dict['add_connection_small_specie_rate'], 
                                  add_node_specie_rate = configuration_dict['add_node_specie_rate'], activate_connection_rate = configuration_dict['activate_connection_rate'], 
                                  remove_connection_rate = configuration_dict['remove_connection_rate'], children = configuration_dict['children'], 
                                  crossover_rate = configuration_dict['crossover_rate'], saving = configuration_dict['saving'], limiting_species = configuration_dict['limiting_species'], limiting_threshold = configuration_dict['limiting_threshold'], same_fitness_limit = configuration_dict['same_fitness_limit'], keep_parents = configuration_dict['keep_parents'], age_significance = configuration_dict['age_significance'])
        self.alone_training_iterations = alone_training_iterations
        self.max_number_of_games = max_number_of_games
        self.max_number_of_steps = max_number_of_steps
        self.training_public_key = training_public_key
        self.training_private_key = training_private_key
        self.socket1_training_public_key = socket1_training_public_key
        self.socket1_training_private_key = socket1_training_private_key
        self.socket2_training_public_key = socket2_training_public_key
        self.socket2_training_private_key = socket2_training_private_key
        self.socket3_training_public_key = socket3_training_public_key
        self.socket3_training_private_key = socket3_training_private_key
        self.client = None
    
    def alone_training(self):
        list_envs = []
        list_states = []
        list_games = []
        list_steps = []
        list_keep_going = []
        for i in range(self.alone_training_iterations):
            self.neat.set_generation_iter(i)
            number_genomes = self.neat.get_number_of_genomes()
            self.neat.reset_fitnesses()
            length = len(list_envs)
            for j in range(number_genomes):
                if j >= length:
                    env = gym.make(self.gym_game_name)
                    state = env.reset()
                    if type(state) == type(tuple()): 
                        state = state[0]
                    state = gym_http_server.flat_state(state)
                    reward = 1
                    done = False
                    list_envs.append(env)
                    list_states.append(state)
                    list_games.append(0)
                    list_steps.append(0)
                    list_keep_going.append(True)
                else:
                    env = list_envs[j]
                    state = env.reset()
                    if type(state) == type(tuple()): 
                        state = state[0]
                    state = gym_http_server.flat_state(state)
                    reward = 1
                    done = False
                    list_states[j] = state
                    list_games[j] = 0
                    list_steps[j] = 0
                    list_keep_going[j] = True
            
            stop_execution = False
            first_time = True
            while True in list_keep_going[:number_genomes]:
                for j in range(number_genomes):
                    list_steps[j]+=1
                start = time.time()
                out = self.neat.ff_all_genomes(list_states, self.threads_for_alone_training,number_genomes, list_genomes_to_ff = list_keep_going)
                end = time.time()
                if first_time:
                    first_time = False
                    if end-start >= 1:
                        stop_execution = True
                k = 0
                for j in range(number_genomes):
                    if list_keep_going[j]:
                        minimum = out[k][0]
                        action = 0
                        for z in range(len(out[k])):
                            if out[k][z] > minimum:
                                minimum = out[k][z]
                                action = z
                        k+=1
                        ret = list_envs[j].step(action)
                        state = ret[0]
                        reward = ret[1]
                        done = ret[2] 
                        if self.configuration_dict['minimum_reward'] <= 0:
                            reward += -1*self.configuration_dict['minimum_reward']
                        if self.configuration_dict['minimum_reward'] < 1:
                            reward+=1
                        state = gym_http_server.flat_state(state)
                        list_states[j] = state
                        self.neat.increment_fitness_of_genome_ith(j,reward/self.max_number_of_games)
                        if done or list_steps[j] >= self.max_number_of_steps:
                            list_steps[j] = 0
                            list_games[j]+=1
                            state = list_envs[j].reset()
                            if type(state) == type(tuple()): 
                                state = state[0]
                            state = gym_http_server.flat_state(state)
                            list_states[j] = state
                        if list_games[j] >= self.max_number_of_games:
                            list_keep_going[j] = False
            self.neat.generation_run()
            if stop_execution:
                self.alone_training_iterations = i+1
                break
        self.neat.reset_fitnesses()
    
    def distributed_training(self, remote_ip, remote_port,genomes_per_client,max_number_of_trainers, ip = '127.0.0.1', port=5000, timeout=3):
        if genomes_per_client <= 0 or max_number_of_trainers <= 0:
            print("Error, genomes per client can't be <= 0, same for max number of trainers")
            exit(1)

        link = self.socket1_training_public_key+self.socket2_training_public_key+self.socket3_training_public_key

        
        # checking with http api if this training already exists , in that case wait 10 secs and re ask
        # if this training keeps existing for more than 5 minutes close the process
        
        # setting import parameters
        gym_http_server.glob_val.max_number_of_trainers = max_number_of_trainers
        gym_http_server.glob_val.max_number_genomes_per_client = genomes_per_client
        gym_http_server.glob_val.max_number_of_steps = self.max_number_of_steps
        gym_http_server.glob_val.max_number_of_games = self.max_number_of_games
        gym_http_server.glob_val.minimum_reward = self.configuration_dict['minimum_reward']
        gym_http_server.glob_val.sub_games = self.configuration_dict['sub_games']
        gym_http_server.glob_val.env_id = self.gym_game_name
        # starting the gym server on another thread
        gym_http_server.init_gym_server(self.training_private_key,ip,port)
        # starting the timeout for the environments
        gym_http_server.init_environments_timeout(self.training_private_key, timeout, self.t_val)
        
        # initializing the host client with ezspark proxy
        self.client = ezclient.Client(self.training_public_key,training_private_key = self.training_private_key,neat_class = self.neat, gym_game_name = self.gym_game_name, buffer_size = 100000, genome_input = self.input_size, genome_output = self.output_size, url_name = link)
        # connection for p2p through ezprotocol
        
        ret = {}
        
        while not 'connected' in ret or ret['connected']:
            ret = requests.get('https://alpha-p2p.ezspark.ai/rest/isTrainingConnected/'+self.training_private_key, verify=False)
            ret = json.loads(ret.content)
            if 'connected' in ret and not ret['connected']:
                break
            time.sleep(2)
        
        self.client.connect(remote_ip,remote_port, genomes_per_client = genomes_per_client)
        self.neat.set_generation_iter(self.alone_training_iterations)
        gym_http_server.glob_val.generation = self.alone_training_iterations
        generation = self.alone_training_iterations
        gym_http_server.glob_val.enter_critical_section()
        gym_http_server.glob_val.set_generation(generation)
        gym_http_server.glob_val.set_genomes(self.neat.get_number_of_genomes())
        gym_http_server.glob_val.exit_critical_section()
        
        socket1_thread = postTunnelRun('http://'+ip+':'+str(port),self.socket1_training_public_key, self.socket1_training_private_key,
                                        self.neat, self.gym_game_name, 100000, self.input_size, self.output_size, link, remote_ip, remote_port, genomes_per_client)
        socket2_thread = postTunnelRun('http://'+ip+':'+str(port),self.socket2_training_public_key, self.socket2_training_private_key,
                                        self.neat, self.gym_game_name, 100000, self.input_size, self.output_size, link, remote_ip, remote_port, genomes_per_client)
        socket3_thread = postTunnelRun('http://'+ip+':'+str(port),self.socket3_training_public_key, self.socket3_training_private_key,
                                        self.neat, self.gym_game_name, 100000, self.input_size, self.output_size, link, remote_ip, remote_port, genomes_per_client)
        
        socket1_thread.start()
        socket2_thread.start()
        socket3_thread.start()
        ss = 0
        while(True):
            # keep the communication active
            while(not self.client.is_disconnected()):
                self.client.host_direct_main_loop()
            # we have been disconnected by the server (bad requests or time out)
            if self.client.got_broken_pipe():
                ss+=1
                ret = {}
                # let's check if the problem is that someone else already started this host training
                self.client.connect(remote_ip, remote_port, genomes_per_client = genomes_per_client)
                continue
            ss = 0
            #blocking the timeouts that must talk with the p2p
            gym_http_server.glob_val.enter_critical_section()
            gym_http_server.glob_val.timeout_flag = True
            gym_http_server.glob_val.exit_critical_section()
            # we are here, a trainer is communicating with us
            is_ok = True
            current_id = None
            we_care = False
            flag2 = False
            # if we did already assigned an identifier to him:
            if self.client.trainer_has_identifier():
                # instantiate the identifier assigned
                current_id = self.client.get_identifier().decode('utf-8')
                current_id_index = self.client.get_index_value(current_id)
                gym_http_server.glob_val.enter_critical_section()
                environment_name = self.client.get_instance_name().decode('utf-8')
                
                # if environment is in reverse: we didn't finished the training for this environment name. or if the p2p id or the number of fitnesses do not match, it is malicious.
                # For the p2p id case it could more likely happen during http requests (for timeout), in those case the trainer will reset itself and restart from scratch without conencting with us
                # but if we reach this point either we were unlucky and we will close and restart the connection, or is really malicious
                if environment_name in gym_http_server.glob_val.reverse_shared_d or current_id not in gym_http_server.glob_val.id_p2p or self.client.get_number_of_fitnesses() != gym_http_server.glob_val.id_p2p[current_id]['n_genomes']:
                    is_ok = False
                    gym_http_server.glob_val.close_environments(environment_name)
                # this environment name could have been closed by timeout, ok do not disconnect lets just assign new genomes.
                # the other case is when it is malicious, well communicate genomes to him anyway
                elif environment_name not in gym_http_server.glob_val.checking_states:
                    flag2 = True
                else:#everything was fine
                    # ok assign the right fitnesses
                    d = gym_http_server.glob_val.checking_states[environment_name]['rewards']
                    l = list(d.keys())
                    l.sort()
                    l_rew = []
                    for i in l:
                        l_rew.append(d[i]/self.max_number_of_games)
                    self.client.set_fitnesses(l_rew)
                # if it is not malicious, or we could care about this
                if is_ok and not flag2:
                    #lets first surely check if we really care
                    if self.client.we_do_care_about_this_trainer():
                        we_care = True
                        try:
                            l = gym_http_server.glob_val.checking_states[environment_name]['list_to_check']
                        except:
                            l = []
                        if l == []:
                            is_ok = False
                        else:
                            gym_http_server.glob_val.exit_critical_section()# we exit here becuase it could take a little bit of time
                            is_ok = self.client.comparing_history_is_ok(l,0.00001)# comparing the history
                            gym_http_server.glob_val.enter_critical_section()
                    gym_http_server.glob_val.checking_states.pop(environment_name, None)
                gym_http_server.glob_val.exit_critical_section()
            if is_ok:# is ok as trainer, or is malicious but did not relly hurted us, or is ok but timeout timed out it
                if self.client.set_body(1, gym_http_server.glob_val.current_index, gym_http_server.glob_val.next_index, flag2):# if is true, a generation run has been completed
                    #setting some informations for the http retrivieng information host side
                    generation+=1
                    gym_http_server.glob_val.enter_critical_section()
                    gym_http_server.glob_val.set_generation(generation)
                    gym_http_server.glob_val.set_genomes(self.neat.get_number_of_genomes())
                    gym_http_server.glob_val.exit_critical_section()
                    if current_id == None:
                        current_id = self.client.get_identifier().decode('utf-8')
                    gen_run = self.neat.get_generation_iter()
                    if gen_run >= self.configuration_dict['generations']:
                        print("end of the training")
                        exit(0)
                    gym_http_server.glob_val.enter_critical_section()
                    gym_http_server.glob_val.generation = generation
                    gym_http_server.glob_val.close_from_timeouts()
                    gym_http_server.glob_val.current_index = gym_http_server.glob_val.next_index
                    gym_http_server.glob_val.next_index = gym_http_server.glob_val.generate_indexing()
                    gym_http_server.glob_val.exit_critical_section()
                else:
                    if current_id == None:
                        current_id = self.client.get_identifier().decode('utf-8')
                    else:
                        if we_care:
                            gym_http_server.glob_val.enter_critical_section()
                            gym_http_server.glob_val.close_from_timeouts(id = current_id_index)
                            gym_http_server.glob_val.exit_critical_section()
                            
                if current_id == None:
                    current_id = self.client.get_identifier().decode('utf-8')
                current_id_index = self.client.get_index_value(current_id)
                gym_http_server.glob_val.enter_critical_section()
                gym_http_server.glob_val.id_p2p[current_id] = {'n_genomes':self.client.get_trainer_n_genomes(),'index':current_id_index}
                gym_http_server.glob_val.exit_critical_section()
            else:
                if current_id != None:
                    if current_id in gym_http_server.glob_val.id_p2p:
                        gym_http_server.glob_val.enter_critical_section()
                        gym_http_server.glob_val.id_p2p.pop(current_id, None)
                        gym_http_server.glob_val.exit_critical_section()
                self.client.set_body(0, gym_http_server.glob_val.current_index, gym_http_server.glob_val.next_index, False)# we set the body to tell the server to close this trainer
            # unlocking timeouts
            gym_http_server.glob_val.enter_critical_section()
            gym_http_server.glob_val.timeout_flag = False
            gym_http_server.glob_val.exit_critical_section()



from pyezspark import protocol_host as host
from pyezspark import protocol_trainer as trainer 
import requests
import json
import gym
import warnings

class EzSpark:
    def __init__(self, training_public_key, training_private_key = None, max_number_of_genomes_per_client = 100, t_val = 5, max_number_of_trainers = 10):
        warnings.filterwarnings('ignore')
        self.API_BASE_URL = 'https://api.ezspark.ai/trainings/'
        self.training_public_key = training_public_key
        self.training_private_key = training_private_key
        self.remote_ip = '62.112.9.80'
        self.remote_port = 3389
        self.max_number_of_genomes_per_client = max_number_of_genomes_per_client
        self.t_val = t_val
        self.max_number_of_trainers = max_number_of_trainers
        if self.max_number_of_genomes_per_client < 1:
            print("max number of genomes must be >= 1")
            exit(1)
        
    def execute(self, username = 'erre4', threads = 4):
        try:
            if self.training_private_key == None:
                # trainer
                ret = requests.get(self.API_BASE_URL+self.training_public_key)
                ret = json.loads(ret.text)
                gym_game_name = ret['gym_game_name']
                input_size = ret['input_size']
                output_size = ret['output_size']
                max_number_of_games = ret['max_number_of_games']
                max_number_of_steps = ret['max_number_of_steps']
                sub_games = ret['sub_games']
                t = trainer.Trainer(gym_game_name, self.training_public_key, input_size, output_size, max_number_of_games, max_number_of_steps, threads, username = username, sub_games = sub_games)
                t.connect(self.remote_ip, self.remote_port)
                t.train()
            else:
                # host
                headers = {'authorization':'private_key '+self.training_private_key}
                ret = requests.get(self.API_BASE_URL+self.training_public_key, headers = headers)
                ret = json.loads(ret.text)
                gym_game_name = ret['gym_game_name']
                max_number_of_games = ret['max_number_of_games']
                max_number_of_steps = ret['max_number_of_steps']
                alone_training_iterations = ret['alone_training_iterations']
                training_public_key1 = ret['training_public_key1']
                training_public_key2 = ret['training_public_key2']
                training_public_key3 = ret['training_public_key3']
                training_private_key1 = ret['training_private_key1']
                training_private_key2 = ret['training_private_key2']
                training_private_key3 = ret['training_private_key3']
                
                #check the
                # 1 game must exists
                # 2 action space must be discrete
                # 3 action space must be a list of lists or a numpy array
                
                try:
                    env = gym.make(gym_game_name)
                    env.close()
                
                except:
                    print("the game with name "+gym_game_name+" does not exist")
                    exit(1)
                
                h = host.Host(gym_game_name, alone_training_iterations,ret,max_number_of_games,max_number_of_steps,self.training_public_key,self.training_private_key,
                training_public_key1,training_private_key1,training_public_key2,training_private_key2,training_public_key3,
                training_private_key3, self.t_val)
                h.alone_training()
                h.distributed_training(self.remote_ip, self.remote_port,self.max_number_of_genomes_per_client,self.max_number_of_trainers)
        except:
            print("something went wrong")
            exit(1)



import threading, time, random
import uuid
import gym
import numpy as np
import six
import argparse
import sys
import json
import string
import logging
from datetime import datetime, timedelta
import requests
from http.server import HTTPServer, BaseHTTPRequestHandler

logger = logging.getLogger('werkzeug')
logger.setLevel(logging.ERROR)
timeout_value = 5

def flat_state(vector):
    if type(vector) == np.ndarray:
        return vector.flatten().tolist()
    elif type(vector) == list:
        v = vector
        if any(isinstance(i, list) for i in vector):
            v = list(chain(*vector))
        return v
    return None

init_all = 0

########## Globals that must be modified and shared by threads ##########
class GlobalVal:
    def __init__(self, max_number_genomes_per_client, max_number_trainers, max_number_of_steps, max_number_of_games, t_val):
        if max_number_genomes_per_client <= 0:
            print("Error, max number of genomes can't be less than 1")
            exit(1)
        if max_number_trainers <= 0:
            print("Error, max number trainers can't be less than 1")
            exit(1)
        if max_number_of_steps <= 0 or max_number_of_games <= 0:
            print("Error, max number of steps and games can't be less than 1")
            exit(1)
        self.mutex = threading.Lock()
        self.max_number_genomes_per_client = max_number_genomes_per_client#max number of genomes per clients
        self.max_number_trainers = max_number_trainers#max number of trainers that we can allow togheter
        self.max_number_of_games = max_number_of_games
        self.max_number_of_steps = max_number_of_steps
        self.minimum_reward = 0
        self.sub_games = 1
        self.env_id = None
        self.generation = 1
        self.id_p2p = {}#each p2p id is associated to n_genomes
        self.shared_d = {}
        self.timeout_d = {}#handled by other thread
        self.timeout_flag = False;
        self.training_private_key = None
        self.generation = 0
        self.total_number_of_genomes = 0
        self.t_val = t_val
        self.stored_envs = []
        self.stored_envs_flag = False
        '''
        shared_d struct:
        (key, value) = (random id,{'ids':[list_of_environment_ids], 'interactions':integer})
        '''
        self.reverse_shared_d = {}
        '''
        reverse_shared_d struct:
        (key, value) = (environment_id,[shared_d_id,n_games,n_steps,stopped_training (True/false), last call was end of game in its instance (True/false)])
        '''
        self.checking_states = {}
        '''
        checking_states struct:
        (key, value) = (environment_id[0] from the list of shared_d, {'index':0, 'associated_id':id, 'current_index':self.current_index-1,'index_got':self.current_index, 'length': n, 'list_to_check':[]})
        '''
        self.current_trainers = 0
        self.current_index = self.generate_indexing()
        self.next_index = self.generate_indexing()
    
        
    
    def set_private_key(self,private_key):
        self.training_private_key = private_key
    
    def set_generation(self,generation):
        self.generation = generation
    
    def set_genomes(self,genomes):
        self.total_number_of_genomes = genomes
        
    def generate_indexing(self):
        return 1 + int(random.uniform(0.8,1)*self.max_number_genomes_per_client/2)
    
    def enter_critical_section(self):
        self.mutex.acquire()
    
    def exit_critical_section(self):
        self.mutex.release()
    
    def generate_id(self,length):
        characters = string.ascii_letters + string.digits + string.punctuation
        id = ''.join(random.choice(characters) for i in range(length))
        return id.replace("'",",").replace('"','!')
    
    def create_environments_is_ok(self, n_instances):
        if n_instances < 1 or n_instances > self.max_number_genomes_per_client*self.sub_games:
            return False
        if self.current_trainers >= self.max_number_trainers:
            return False
        return True
    
    def check_states(self,list_of_environments_id, states, reward):
        n = len(list_of_environments_id)
        id = ''
        for i in range(n):
            if list_of_environments_id[i] not in self.reverse_shared_d:
                return False
            else:
                id = self.reverse_shared_d[list_of_environments_id[i]][0]
                break
        list_of_state_ids = list(self.checking_states.keys())
        for i in range(len(list_of_state_ids)):
            if self.checking_states[list_of_state_ids[i]]['associated_id'] == id:
                for j in range(n):
                    self.checking_states[list_of_state_ids[i]]['rewards'][self.reverse_shared_d[list_of_environments_id[j]][5]] += reward[j]
                    # if it enters in this if we are recording stuff
                    # remember we must record also the actions given back
                    # and this will always happen!
                    if self.checking_states[list_of_state_ids[i]]['index'] == self.checking_states[list_of_state_ids[i]]['current_index']:
                        l = flat_state(states[j])
                        for kk in l:
                            self.checking_states[list_of_state_ids[i]]['list_to_check'].append(kk)
                        self.checking_states[list_of_state_ids[i]]['current_index']+=self.checking_states[list_of_state_ids[i]]['index_got']
                        self.checking_states[list_of_state_ids[i]]['current_index'] = self.checking_states[list_of_state_ids[i]]['current_index']%self.checking_states[list_of_state_ids[i]]['length']
                    self.checking_states[list_of_state_ids[i]]['index']+=1
                    self.checking_states[list_of_state_ids[i]]['index'] = self.checking_states[list_of_state_ids[i]]['index']%self.checking_states[list_of_state_ids[i]]['length']
                return True
        return False
    # the history take into account also the actions given back             
    def add_actions(self, list_of_environments_id, actions):
        list_of_environments_id, actions = zip(*sorted(zip(list_of_environments_id, actions)))
        n = len(list_of_environments_id)
        id = ''
        for i in range(n):
            if list_of_environments_id[i] not in self.reverse_shared_d:
                return False
            else:
                id = self.reverse_shared_d[list_of_environments_id[i]][0]
                break
        list_of_state_ids = list(self.checking_states.keys())
        for i in range(len(list_of_state_ids)):
            if self.checking_states[list_of_state_ids[i]]['associated_id'] == id:
                for j in actions:
                    if self.checking_states[list_of_state_ids[i]]['index'] == self.checking_states[list_of_state_ids[i]]['current_index']:
                        self.checking_states[list_of_state_ids[i]]['list_to_check'].append(float(j))
                        self.checking_states[list_of_state_ids[i]]['current_index']+=self.checking_states[list_of_state_ids[i]]['index_got']
                        self.checking_states[list_of_state_ids[i]]['current_index'] = self.checking_states[list_of_state_ids[i]]['current_index']%self.checking_states[list_of_state_ids[i]]['length']
                    self.checking_states[list_of_state_ids[i]]['index']+=1
                    self.checking_states[list_of_state_ids[i]]['index'] = self.checking_states[list_of_state_ids[i]]['index']%self.checking_states[list_of_state_ids[i]]['length']
        
    def hash_enviroments(self, list_of_environments_id, states, rewards, current_id):
        '''
        it s asking us to hash these environments, this means that, first these environments
        have been just created, no other environments with same ids should exist, also,
        there should no be more or equals current trainers as max number of trainers. Also,
        the number of environments should not exceeds the number of maximum genomes per client
        '''
        if len(list_of_environments_id) > self.max_number_genomes_per_client*self.sub_games:
            return False
        if self.current_trainers >= self.max_number_trainers:
            return False
        n = len(list_of_environments_id)
        for i in range(n):
            if list_of_environments_id[i] in self.reverse_shared_d:
                return False
        self.current_trainers+=1# adding a trainer 
        id = self.generate_id(32)
        list_prim_id = []
        
        # to hash the rewards
        for i in range(int(n/self.sub_games)):
            list_prim_id.append(self.generate_id(32))
        list_prim_id.sort()
        
        #number of games each copy of the genomes must play
        list_of_iterations = [0]*self.sub_games
        count = 0
        while count != self.max_number_of_games:
            for i in range(self.sub_games):
                list_of_iterations[i]+=1
                count+=1
                if count == self.max_number_of_games:
                    break
        
        self.shared_d[id] = {'ids':list(list_of_environments_id), 'interactions':1}#to an identifier is associated the list
        self.checking_states[list_of_environments_id[0]] = {'index':0, 'associated_id':id, 'current_index':(self.current_index-1)%n,'index_got':self.current_index, 'length': n, 'list_to_check':[], 'rewards':{}}
        for i in range(n):
            self.reverse_shared_d[list_of_environments_id[i]] = [id,0,1,False,False, list_prim_id[i%(int(n/self.sub_games))], list_of_iterations[int(i/int(n/self.sub_games))]]#the reverse, id, games, steps, done or not, last time we called this it was done or not, the id of the genome associated
            self.checking_states[list_of_environments_id[0]]['rewards'][list_prim_id[i%(int(n/self.sub_games))]] = 0
        self.timeout_d[id] = {'current_id_index': self.id_p2p[current_id]['index'],'current_id':current_id, 'shared_d':id, 'checking_states':list_of_environments_id[0], 'reverse_shared_d':list(list_of_environments_id), 'last_interaction':datetime.now(), 'seconds_timeout':0.6+n*(self.generation*0.001+0.2)}
        return self.check_states(list_of_environments_id,states,rewards)
        
    # someone is asking us to make a new step, lets check if its request is fair
    def steps_check(self, list_of_environments_id):
        n = len(list_of_environments_id)
        id = None
        for i in range(n):
            if list_of_environments_id[i] not in self.reverse_shared_d:
                return False
            if id == None:
                id = self.reverse_shared_d[list_of_environments_id[i]][0]
            else:
                if self.reverse_shared_d[list_of_environments_id[i]][0] != id:
                    return False
            if self.reverse_shared_d[list_of_environments_id[i]][3]:#it s done
                return False
        if id == None:
            return False
       
        try:
            n2 = len(self.shared_d[id]['ids'])
            for i in range(n2):
                if self.shared_d[id]['ids'][i] not in list_of_environments_id:
                    return False 
        except:
            return False
        return True
    
    def get_t_val(self):
        return self.t_val
        
    def set_t_val(self, value):
        self.t_val = value
        
    def update_steps(self, list_of_environments_id, done, states, rewards):
        '''
        we must first check that these environments exist
        and that are a sublist of the list of all the environments of the identifier associated to them.
        than we must check that is asking us the environments that are not done yet with the training
        '''
        
        n = len(list_of_environments_id)
        id = None
        for i in range(n):
            if list_of_environments_id[i] not in self.reverse_shared_d:
                return False
            if id == None:
                id = self.reverse_shared_d[list_of_environments_id[i]][0]
            else:
                if self.reverse_shared_d[list_of_environments_id[i]][0] != id:
                    return False
            if self.reverse_shared_d[list_of_environments_id[i]][3]:#it s done
                return False
        if id == None:
            return False
        
        # resetting the last interaction time
        flag = False
        for i in self.timeout_d:
            for j in self.timeout_d[i]['reverse_shared_d']:
                
                if j in list_of_environments_id:
                    self.timeout_d[i]['last_interaction'] = datetime.now()
                    flag = True
                    break
            if flag:
                break
        
        # saving the correct states
        list_of_environments_id, done, states, rewards = zip(*sorted(zip(list_of_environments_id, done, states, rewards)))
        glob_val.check_states(list_of_environments_id,states,rewards)
        for i in range(n):
            self.reverse_shared_d[list_of_environments_id[i]][2]+=1
            if done[i] or self.reverse_shared_d[list_of_environments_id[i]][2] >= self.max_number_of_steps:
                self.reverse_shared_d[list_of_environments_id[i]][2] = 0
                self.reverse_shared_d[list_of_environments_id[i]][1]+=1
                self.reverse_shared_d[list_of_environments_id[i]][4] = True
            else:
                self.reverse_shared_d[list_of_environments_id[i]][4] = False
            if self.reverse_shared_d[list_of_environments_id[i]][1] >= self.reverse_shared_d[list_of_environments_id[i]][6]:
                self.reverse_shared_d[list_of_environments_id[i]][3] = True
        return True
    
    def shutdown_envs(self, list_of_environments_id):
        n = len(list_of_environments_id)
        id = None
        for i in range(n):
            if list_of_environments_id[i] not in self.reverse_shared_d:
                return None
            if id == None:
                id = self.reverse_shared_d[list_of_environments_id[i]][0]
            else:
                if self.reverse_shared_d[list_of_environments_id[i]][0] != id:
                    return None
        if id == None:
            return None
        l = []
        for i in range(n):
            if self.reverse_shared_d[list_of_environments_id[i]][3]:
                l.append(list_of_environments_id[i])
                self.shared_d[id]['ids'].remove(list_of_environments_id[i])
                envs.env_close(list_of_environments_id[i])
                self.reverse_shared_d.pop(list_of_environments_id[i], None)
        if len(self.shared_d[id]['ids']) == 0:
            self.shared_d.pop(id,None)
            self.current_trainers-=1
        return l
    
    def close_environments(self, environment):
        id = self.checking_states[environment]['associated_id']
        self.checking_states.pop(environment,None)
        l = []
        for i in self.reverse_shared_d:
            if self.reverse_shared_d[i][0] == id:
                l.append(i)
        for i in l:
            envs.env_close(i)
            self.reverse_shared_d.pop(i,None)
        self.shared_d.pop(id,None)
        self.timeout_d.pop(id,None)
        self.current_trainers-=1
    
    def close_from_timeouts(self, id = None):
        if id == None:
            l = list(self.timeout_d.keys())
            for i in l:
                if self.timeout_d[i]['reverse_shared_d'][0] in self.checking_states and i not in self.shared_d:#has been already clos only the checking states are still active
                   self.checking_states.pop(self.timeout_d[i]['reverse_shared_d'][0],None)
                   self.timeout_d.pop(i,None)
                elif self.timeout_d[i]['reverse_shared_d'][0] not in self.checking_states and i not in self.shared_d:
                    self.timeout_d.pop(i,None)
                else:
                    self.close_environments(self.timeout_d[i]['reverse_shared_d'][0])
        else:
            l = list(self.timeout_d.keys())
            for i in l:
                if self.timeout_d[i]['current_id_index'] == id:
                    if self.timeout_d[i]['reverse_shared_d'][0] in self.checking_states and i not in self.shared_d:#has been already closed only the checking states are still active
                       self.checking_states.pop(self.timeout_d[i]['reverse_shared_d'][0],None)
                       self.timeout_d.pop(i,None)
                    elif self.timeout_d[i]['reverse_shared_d'][0] not in self.checking_states and i not in self.shared_d:
                        self.timeout_d.pop(i,None)
                    else:
                        self.close_environments(self.timeout_d[i]['reverse_shared_d'][0])
                        
    def set_globals(self,max_number_genomes_per_client, max_number_trainers, max_number_of_steps, max_number_of_games, t_val):
        if max_number_genomes_per_client <= 0:
            print("Error, max number of genomes can't be less than 1")
            exit(1)
        if max_number_trainers <= 0:
            print("Error, max number trainers can't be less than 1")
            exit(1)
        if max_number_of_steps <= 0 or max_number_of_games <= 0:
            print("Error, max number of steps and games can't be less than 1")
            exit(1)
        self.mutex = threading.Lock()
        self.max_number_genomes_per_client = max_number_genomes_per_client#max number of genomes per clients
        self.max_number_trainers = max_number_trainers#max number of trainers that we can allow togheter
        self.max_number_of_games = max_number_of_games
        self.max_number_of_steps = max_number_of_steps
        self.t_val = t_val
        
        


########## Container for environments ##########
class Envs(object):
    """
    Container and manager for the environments instantiated
    on this server.

    When a new environment is created, such as with
    envs.create('CartPole-v0'), it is stored under a short
    identifier (such as '3c657dbc'). Future API calls make
    use of this instance_id to identify which environment
    should be manipulated.
    """
    def __init__(self):
        self.envs = {}
        self.id_len = 16 #increasing from 8 to 16 for security reasons

    def _lookup_env(self, instance_id):
        try:
            return self.envs[instance_id]
        except KeyError:
            raise InvalidUsage('Instance_id {} unknown'.format(instance_id))

    def _remove_env(self, instance_id):
        try:
            del self.envs[instance_id]
        except KeyError:
            raise InvalidUsage('Instance_id {} unknown'.format(instance_id))

    def create(self, env_id, pre_made_env = None, seed=None):
        if pre_made_env == None:
            try:
                env = gym.make(env_id)
                if seed:
                    env.seed(seed)
            except gym.error.Error:
                raise InvalidUsage("Attempted to look up malformed environment ID '{}'".format(env_id))
        else:
            env = pre_made_env
        instance_id = str(uuid.uuid4().hex)[:self.id_len]
        self.envs[instance_id] = env
        return instance_id
        
    def create_rough(self, env_id, seed=None):
        try:
            env = gym.make(env_id)
            if seed:
                env.seed(seed)
            return env
        except gym.error.Error:
            raise InvalidUsage("Attempted to look up malformed environment ID '{}'".format(env_id))
            return None

    def list_all(self):
        return dict([(instance_id, env.spec.id) for (instance_id, env) in self.envs.items()])

    def reset(self, instance_id):
        environment = self._lookup_env(instance_id)
        obs = environment.reset()
        return flat_state(obs)

    def step(self, instance_id, action, render):
        env = self._lookup_env(instance_id)
        if isinstance( action, six.integer_types ):
            nice_action = action
        else:
            nice_action = np.array(action)
        if render:
            env.render()
        [observation, reward, done, info] = env.step(nice_action)
        obs_jsonable = flat_state(observation)
        return [obs_jsonable, reward, done, info]

    def get_action_space_contains(self, instance_id, x):
        env = self._lookup_env(instance_id)
        return env.action_space.contains(int(x))

    def get_action_space_info(self, instance_id):
        env = self._lookup_env(instance_id)
        return self._get_space_properties(env.action_space)

    def get_action_space_sample(self, instance_id):
        env = self._lookup_env(instance_id)
        action = env.action_space.sample()
        if isinstance(action, (list, tuple)) or ('numpy' in str(type(action))):
            try:
                action = action.tolist()
            except TypeError:
                print(type(action))
                print('TypeError')
        return action

    def get_observation_space_contains(self, instance_id, j):
        env = self._lookup_env(instance_id)
        info = self._get_space_properties(env.observation_space)
        for key, value in j.items():
            # Convert both values to json for comparibility
            if json.dumps(info[key]) != json.dumps(value):
                print('Values for "{}" do not match. Passed "{}", Observed "{}".'.format(key, value, info[key]))
                return False
        return True

    def get_observation_space_info(self, instance_id):
        env = self._lookup_env(instance_id)
        return self._get_space_properties(env.observation_space)

    def _get_space_properties(self, space):
        info = {}
        info['name'] = space.__class__.__name__
        if info['name'] == 'Discrete':
            info['n'] = space.n
        elif info['name'] == 'Box':
            info['shape'] = space.shape
            # It's not JSON compliant to have Infinity, -Infinity, NaN.
            # Many newer JSON parsers allow it, but many don't. Notably python json
            # module can read and write such floats. So we only here fix "export version",
            # also make it flat.
            info['low']  = [(x if x != -np.inf else -1e100) for x in np.array(space.low ).flatten()]
            info['high'] = [(x if x != +np.inf else +1e100) for x in np.array(space.high).flatten()]
        elif info['name'] == 'HighLow':
            info['num_rows'] = space.num_rows
            info['matrix'] = [((float(x) if x != -np.inf else -1e100) if x != +np.inf else +1e100) for x in np.array(space.matrix).flatten()]
        return info

    def monitor_start(self, instance_id, directory, force, resume, video_callable):
        env = self._lookup_env(instance_id)
        if video_callable == False:
            v_c = lambda count: False
        else:
            v_c = lambda count: count % video_callable == 0
        self.envs[instance_id] = gym.wrappers.Monitor(env, directory, force=force, resume=resume, video_callable=v_c) 

    def monitor_close(self, instance_id):
        env = self._lookup_env(instance_id)
        env.close()

    def env_close(self, instance_id):
        env = self._lookup_env(instance_id)
        env.close()
        self._remove_env(instance_id)
        


########## App setup ##########
envs = Envs()
glob_val = GlobalVal(10,10,5,1,5)

########## Error handling ##########
class InvalidUsage(Exception):
    status_code = 400
    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv

def get_required_param(json, param):
    if json is None:
        logger.info("Request is not a valid json")
        raise InvalidUsage("Request is not a valid json")
    value = json.get(param, None)
    if (value is None) or (value=='') or (value==[]):
        logger.info("A required request parameter '{}' had value {}".format(param, value))
        raise InvalidUsage("A required request parameter '{}' was not provided".format(param))
    return value

def get_optional_param(json, param, default):
    if json is None:
        logger.info("Request is not a valid json")
        raise InvalidUsage("Request is not a valid json")
    value = json.get(param, None)
    if (value is None) or (value=='') or (value==[]):
        logger.info("An optional request parameter '{}' had value {} and was replaced with default value {}".format(param, value, default))
        value = default
    return value


def generateTable():
    timeout = '1000s'
    date = datetime.now()
    html_tag = '<html>'
    head_tag = '<head>'
    css_link = '<link href="https://fonts.googleapis.com/css?family=Lato" rel="stylesheet"><link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/meyer-reset/2.0/reset.min.css">'
    style = '<style> body { font-family: "lato", sans-serif; } .container { max-width: 1000px; margin-left: auto; margin-right: auto; padding-left: 10px; padding-right: 10px; } h2 { font-size: 26px; margin: 20px 0; text-align: center; } h2 small { font-size: 0.5em; } .responsive-table li { border-radius: 3px; padding: 25px 30px; display: flex; justify-content: space-between; margin-bottom: 25px; } .responsive-table .table-header { background-color: #95A5A6; font-size: 14px; text-transform: uppercase; letter-spacing: 0.03em; } .responsive-table .table-row { background-color: #ffffff; box-shadow: 0px 0px 9px 0px rgba(0, 0, 0, 0.1); } .responsive-table .col-1 { flex-basis: 30%; } .responsive-table .col-2 { flex-basis: 30%; } .responsive-table .col-3 { flex-basis: 10%; } .responsive-table .col-4 { flex-basis: 10%; } .responsive-table .col-5 { flex-basis: 10%; } .responsive-table .col-6 { flex-basis: 10%; } @media all and (max-width: 767px) { .responsive-table .table-header { display: none; } .responsive-table li { display: block; } .responsive-table .col { flex-basis: 100%; } .responsive-table .col { display: flex; padding: 10px 0; } .responsive-table .col:before { color: #6C7A89; padding-right: 10px; content: attr(data-label); flex-basis: 50%; text-align: right; } } </style> <script> window.console = window.console || function(t) {}; </script> <script> if (document.location.search.match(/type=embed/gi)) { window.parent.postMessage("resize", "*"); } </script>'
    end_head_tag = '</head>'
    body = '<body translate="no">'
    title = '<div class="container"><h2>'+glob_val.training_private_key+' Training</h2>'
    number_of_trainers = '<div><b>Number of Current Trainers: </b>'+str(glob_val.current_trainers)+'</div>'
    generation = '<div><b>Generation: </b>'+str(glob_val.generation)+'</div>'
    total_genomes = '<div><b>Number of genomes for this generation: </b>'+str(glob_val.total_number_of_genomes)+'</div>'
    table_tag = '<ul class="responsive-table">'
    table_head = '<li class="table-header"><div class="col col-1">Peer Id</div><div class="col col-2">Timeout</div><div class="col col-3">N Genomes</div><div class="col col-4">Ended Games</div><div class="col col-5">Games Playing</div><div class="col col-6">Genome index</div></li>'
    table_body = ''
    l = list(glob_val.timeout_d.keys())
    for i in l:
        if glob_val.timeout_d[i]['checking_states'] in glob_val.checking_states: 
            gym_id = i
            p2p_id = glob_val.timeout_d[i]['current_id']
            n_genomes = glob_val.checking_states[glob_val.timeout_d[i]['checking_states']]['length']
            ended_games = n_genomes
            games_playing = 0
            gym_id = gym_id.replace('<','')
            gym_id = gym_id.replace('>','')
            p2p_id = gym_id.replace('>','')
            p2p_id = gym_id.replace('<','')
            
            if i in glob_val.shared_d:
                games_playing = len(glob_val.shared_d[i]['ids'])
                ended_games = n_genomes-games_playing
            genome_index = glob_val.timeout_d[i]['current_id_index']
            
            if glob_val.timeout_d[i]['checking_states'] in glob_val.checking_states and i not in glob_val.shared_d:#has been already closed only the checking states are still active
                if glob_val.timeout_flag:
                    timeout = 'is in pause'
                else:
                    if (date-glob_val.timeout_d[i]['last_interaction']).total_seconds() >= (glob_val.current_trainers+1)*15*glob_val.timeout_d[i]['seconds_timeout']:
                        timeout = 'is expiring'
                    else:
                        timeout = str((glob_val.current_trainers+1)*15*glob_val.timeout_d[i]['seconds_timeout'] + timeout_value - ((date-glob_val.timeout_d[i]['last_interaction']).total_seconds()))+'s'
                        
            elif glob_val.timeout_d[i]['checking_states'] not in glob_val.checking_states and i not in glob_val.shared_d:
                timeout = 'is expiring'
            else:
                if (date-glob_val.timeout_d[i]['last_interaction']).total_seconds() >= (glob_val.current_trainers+1)*3*glob_val.timeout_d[i]['seconds_timeout']:
                    timeout = 'is expiring'
                else:
                    timeout = str((glob_val.current_trainers+1)*3*glob_val.timeout_d[i]['seconds_timeout'] + timeout_value - ((date-glob_val.timeout_d[i]['last_interaction']).total_seconds()))+'s'
            
            
            table_body+='<li class="table-row"><div class="col col-1" data-label="P2P id">'+str(p2p_id)+'</div><div class="col col-2" data-label="Gym Id">'+str(timeout)+'</div><div class="col col-3" data-label="N Genomes">'+str(n_genomes)+'</div><div class="col col-4" data-label="Ended Games">'+str(ended_games)+'</div><div class="col col-5" data-label="Games Playing">'+str(games_playing)+'</div><div class="col col-6" data-label="Genome index">'+str(genome_index)+'</div></li>'
    end_table_tag = '</ul></div>'
    end_body_tag = '</body>'
    end_html_tag = '</html>'
    return html_tag+head_tag+css_link+style+end_head_tag+body+title+number_of_trainers+generation+total_genomes+table_tag+table_head+table_body+end_table_tag+end_body_tag+end_html_tag


def env_create(js):
    """
    Create an instance of the specified environment

    Parameters:
        - env_id: gym environment ID string, such as 'CartPole-v0'
        - n_instances: number of instances to instantiate
        - identifier
    Returns:
        - instance_id: {obs, reward}, ...
    """
    glob_val.enter_critical_section()
    if glob_val.current_trainers >= glob_val.max_number_trainers:
        glob_val.exit_critical_section()
        ret = {}
        ret['full'] = True
        return str(ret)
    env_id = get_required_param(js, 'env_id')
    n_instances = get_required_param(js, 'n_instances')
    identifier = get_required_param(js,'identifier')
    seed = None
    if type(n_instances) != int or type(env_id) != str or not glob_val.create_environments_is_ok(n_instances) or identifier not in glob_val.id_p2p or identifier in glob_val.id_p2p and glob_val.id_p2p[identifier]['n_genomes']*glob_val.sub_games != n_instances:
        glob_val.exit_critical_section()
        ret = {}
        ret['ok'] = False
        return str(ret)
    ret = {}
    l1 = []
    l2 =  []
    l3 = []
    pre_made_envs = None
    if glob_val.stored_envs_flag:
        pre_made_envs = glob_val.stored_envs
        glob_val.stored_envs = []
        glob_val.stored_envs_flag = False
    glob_val.exit_critical_section()
    for i in range(n_instances):
        pre_made = None
        if pre_made_envs != None:
            pre_made = pre_made_envs[i]
        instance = envs.create(env_id, pre_made, seed)
        l1.append(instance)
        e = envs.reset(l1[i])
        if type(e) == type(tuple()):
            e = e[0]
        e = flat_state(e)
        reward = 1
        if glob_val.minimum_reward <= 0:
            reward = -1*glob_val.minimum_reward
        if glob_val.minimum_reward < 1:
            reward += 1
        ret[instance] = e
        l2.append(ret[instance])
        l3.append(reward)
    if pre_made_envs != None and n_instances < len(pre_made_envs):
        for i in range(n_instances,len(pre_made_envs)):
            pre_made_envs[i].close()
        
    glob_val.enter_critical_section()
    l1, l2, l3 = zip(*sorted(zip(l1, l2, l3)))
    glob_val.hash_enviroments(l1,l2,l3, identifier)
    glob_val.shutdown_envs(l1)
    glob_val.exit_critical_section()
    return str(ret)

def multi_step(js):
    """ 
    request a step
    returning for each instance id as key a observation, reward, done and reset observation in case
    - instance1:action
    - instance2:action
    - ...
    - instancen:action
    response:
    - instance1:[obs, reward, done, rest_obs]
    - ...
    """
    
    response = {}
    keys = list(js.keys())
    keys.sort()
    glob_val.enter_critical_section()
    if not glob_val.steps_check(keys):
        glob_val.exit_critical_section()
        ret = {}
        ret['ok'] = False
        return str(ret)
    glob_val.exit_critical_section()
    try:
        l1 = []
        l2 = []
        l3 = []
        l4 = []
        l5 = []
        l6 = []
        for key in js:
            l5.append(key)
            l6.append(float(js[key]))
            if glob_val.reverse_shared_d[key][4]:
                obs_jsonable = envs.reset(key)
                if type(obs_jsonable) == type(tuple()):
                    obs_jsonable = obs_jsonable[0]
                reward = 1
                if glob_val.minimum_reward <= 0:
                    reward = -1*glob_val.minimum_reward
                if glob_val.minimum_reward < 1:
                    reward += 1
                done = False
                
            else:
                e = envs.step(key, js[key], False)
                obs_jsonable = e[0]
                reward = e[1]
                if glob_val.minimum_reward <= 0:
                    reward += -1*glob_val.minimum_reward
                if glob_val.minimum_reward < 1:
                    reward += 1
                done = e[2]
            obs_jsonable = flat_state(obs_jsonable)
            response[key] = [obs_jsonable, done]
            l1.append(key)
            l2.append(obs_jsonable)
            l3.append(reward)
            l4.append(done)
        l1, l2, l3, l4 = zip(*sorted(zip(l1, l2, l3, l4)))
    
    except:
        ret = {}
        ret['ok'] = False
        return str(ret)
       
    glob_val.enter_critical_section()
    try:
        glob_val.add_actions(l5,l6)
        glob_val.update_steps(l1,l4,l2,l3)
        glob_val.shutdown_envs(l1)
    
    except:
        ret = {}
        ret['ok'] = False
        glob_val.exit_critical_section()
        return str(ret)

    glob_val.exit_critical_section()
    return str(response)

def get_status():
    glob_val.enter_critical_section()
    s = generateTable()
    glob_val.exit_critical_section()
    return s
    
    
    



class S(BaseHTTPRequestHandler):
    def _set_headers(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def _html(self, message):
        """This just generates an HTML document that includes `message`
        in the body. Override, or re-write this do do more interesting stuff.
        """
        content = f"<html><body><h1>{message}</h1></body></html>"
        return content.encode("utf8")  # NOTE: must return a bytes object!

    def do_GET(self):
        self._set_headers()
        self.wfile.write(bytes(get_status(), encoding = 'utf-8'))

    def do_HEAD(self):
        self._set_headers()


def run(server_class=HTTPServer, handler_class=S, addr="localhost", port=8000):
    server_address = (addr, port)
    httpd = server_class(server_address, handler_class)
    httpd.serve_forever()

class ServerRun(threading.Thread):
    def __init__(self, ip, port):
        threading.Thread.__init__(self)
        self.ip = ip
        self.port = port
 
        # helper function to execute the threads
    def run(self):
        print('starting the gym server on '+self.ip+':'+str(self.port))
        run(addr = self.ip, port = self.port)

class timeoutRun(threading.Thread):
    def __init__(self, timeout, training_private_key):
        threading.Thread.__init__(self)
        self.timeout = timeout
        self.timeout_dict = {}
        self.training_private_key = training_private_key
        self.polling_url = 'https://alpha-p2p.ezspark.ai/rest/hostPolling/'+self.training_private_key
    def run(self):
        
        while True:
            time.sleep(self.timeout)
            
            # polling
            try:
                ret = requests.get(self.polling_url, verify=False)
            except:
                continue
            
            #pre storing some instances
            date = datetime.now()
            glob_val.enter_critical_section()
            if not glob_val.stored_envs_flag:
                glob_val.exit_critical_section()
                for i in range(glob_val.max_number_genomes_per_client*glob_val.sub_games):
                    date_now = datetime.now()
                    if (date_now-date).total_seconds() > self.timeout:
                        try:
                            ret = requests.get(self.polling_url, verify=False)
                            date = date_now
                        except:
                            print("something went wrong polling the server")
                    instance = envs.create_rough(glob_val.env_id)
                    glob_val.stored_envs.append(instance)
                glob_val.enter_critical_section()
                glob_val.stored_envs_flag = True
            
            # checking the timeouts
            date = datetime.now()
            l = list(glob_val.timeout_d.keys())
            for i in l:
                if glob_val.timeout_d[i]['checking_states'][0] in glob_val.checking_states and i not in glob_val.shared_d:#has been already closed only the checking states are still active
                    if glob_val.timeout_flag and not i in self.timeout_dict:
                        self.timeout_dict[i] = (date-glob_val.timeout_d[i]['last_interaction']).total_seconds()
                    else:
                        if i in self.timeout_dict:
                            glob_val.timeout_d[i]['last_interaction'] = date - timedelta(seconds=self.timeout_dict[i])
                        if (date-glob_val.timeout_d[i]['last_interaction']).total_seconds() >= (glob_val.current_trainers+1)*15*glob_val.timeout_d[i]['seconds_timeout'] + timeout_value:
                            glob_val.checking_states.pop(glob_val.timeout_d[i]['checking_states'],None)
                            glob_val.timeout_d.pop(i,None)
                elif glob_val.timeout_d[i]['checking_states'] not in glob_val.checking_states and i not in glob_val.shared_d:
                    glob_val.timeout_d.pop(i,None)
                else:
                    if (date-glob_val.timeout_d[i]['last_interaction']).total_seconds() >= (glob_val.current_trainers+1)*3*glob_val.timeout_d[i]['seconds_timeout'] + timeout_value:
                        glob_val.close_environments(glob_val.timeout_d[i]['checking_states'])
            if not glob_val.timeout_flag:
                self.timeout_dict = {}
            glob_val.exit_critical_section()

def init_gym_server(private_key,ip = '127.0.0.1', port = 5000):
    glob_val.set_private_key(private_key)
    starting_thread = ServerRun(ip,port)
    starting_thread.start()

def init_environments_timeout(training_private_key, timeout = 3, t_val = 5):
    timeout_value = t_val
    starting_thread = timeoutRun(timeout, training_private_key)
    starting_thread.start()



import setuptools

with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "Pyezspark",
    packages=['pyezspark'],
    version = "2.2.0",
    install_requires=['setuptools>=18.0','wheel','Cython', 'numpy', 'Ezclient', 'requests', 'gym', 'six'],
    author = "Riccardo Viviano",
    author_email = "riccardo.viviano@ezspark.ai",
    description = "Ezspark python package",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/ez-spark/Pyezspark",
    project_urls = {
        "Bug Tracker": "https://github.com/ez-spark/Pyezspark",
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires = ">=3.6"
)



import ezclient
import requests
import time
import json
from datetime import datetime
import random
import string

class Trainer:
    def __init__(self, gym_game_name, training_public_key, input_size, output_size, max_number_of_games, max_number_of_steps, threads, username=None, sub_games = 1):
        if max_number_of_steps <= 0 or max_number_of_games <= 0 or input_size <= 0 or output_size <= 0 or threads <= 0:
            print("Error, input_size and/or output and/or max_number_of_games and/or max_number of steps size are wrong")
            exit(1)
        self.client = ezclient.Client(training_public_key,gym_game_name=gym_game_name, buffer_size = 100000, genome_input = input_size, genome_output = output_size, username=username)
        self.client_secondary = ezclient.Client(training_public_key,gym_game_name=gym_game_name, buffer_size = 100000, genome_input = input_size, genome_output = output_size, username=username)
        self.gym_game_name = gym_game_name
        self.training_public_key = training_public_key
        self.input_size = input_size
        self.output_size = output_size
        self.max_number_of_games = max_number_of_games
        self.max_number_of_steps = max_number_of_steps
        self.sub_games = sub_games
        self.threads = threads
        self.remote_ip = None
        self.remote_port = None
        self.ez_server_addr = 'https://alpha-p2p.ezspark.ai/rest/'
        self.active_training = 'isTrainingActive/'
        self.environment_creation_endpoint = '/v1/envs/'
        self.next_step_endpoint = '/v1/envs/step/'
        self.first_iteration_distributed_training = True
        self.headers = {'Content-type': 'application/json'}
    
    #calling connect with try except after the except re try, but first we ask through http 
    # api if that training exists
    def connect(self, remote_ip, remote_port):
        self.remote_ip = remote_ip
        self.remote_port = remote_port
        self.client.connect(remote_ip,remote_port)
        
    def set_file_descriptor(self,to_secondary):
        if to_secondary:
            fd = self.client.get_file_descriptor()
            self.client_secondary.set_file_descriptor(fd)
        else:
            fd = self.client_secondary.get_file_descriptor()
            self.client.set_file_descriptor(fd)
    
    def send_http_request(self, http_request_string, complete_secondary_training_public_key):
        # the returning value
        id = None
        
        # splitting with all the 3 host sockets public keys
        first_key = complete_secondary_training_public_key[:int(len(complete_secondary_training_public_key)/3)]
        second_key = complete_secondary_training_public_key[int(len(complete_secondary_training_public_key)/3):int(2*len(complete_secondary_training_public_key)/3)]
        third_key = complete_secondary_training_public_key[int(2*len(complete_secondary_training_public_key)/3):]
        # getting the current socket file descriptor
        fd = self.client.get_file_descriptor()
        # init the secondary client
        self.client_secondary.send_http_request_setup(self.remote_ip,self.remote_port,http_request_string,complete_secondary_training_public_key,fd)
        last = datetime.now()
        current = datetime.now()
        limit = 600#10 minutes            
        while True:
            while(not self.client_secondary.is_disconnected()):
                self.client_secondary.trainer_direct_main_loop_http_requests()
                current = datetime.now()
                if (current-last).total_seconds() >= limit:
                    exit(1)
            # if something went wrong
            if self.client_secondary.got_broken_pipe():
                #lets check the training is still active
                
                ret = requests.get(self.ez_server_addr+self.active_training+self.training_public_key, verify = False)
                ret = json.loads(ret.content)
                if not ret['ret']:
                    return None
                
                # try to change the target host socket
                
                temp = first_key
                if self.client_secondary.get_secondary_key() == first_key:
                    temp = second_key
                elif self.client_secondary.get_secondary_key() == second_key:
                    temp = third_key
                self.client_secondary.connect_http_handler(self.remote_ip,self.remote_port,temp,http_request_string,fd)
                continue
            try:
                id = self.client_secondary.get_string().decode('utf-8').replace("'",'"').replace('False','false').replace('True','true')
            except:
                print("Malicious host")
                exit(1)
            break
        return id
    
    def train(self):
        environment_name = None
        last = datetime.now()
        limit = 1200#20 minutes
        while(True):
            self.client.set_sub_games(self.sub_games)
            current = datetime.now()
            if (current-last).total_seconds() >= limit:
                exit(1)
            # communication protocol is acting
            while(not self.client.is_disconnected()):
                self.client.trainer_direct_main_loop()
                current = datetime.now()
                if (current-last).total_seconds() >= limit:
                    exit(1)
            #we have been disconnected by the server (timeout, or we communicated something bad to host, or to the server)
            if self.client.got_broken_pipe():
                #lets check the training is still active
                
                ret = requests.get(self.ez_server_addr+self.active_training+self.training_public_key, verify = False)
                ret = json.loads(ret.content)
                if not ret['ret']:
                    return
                
                
                self.client.connect(self.remote_ip,self.remote_port)
                # adding request to server with http when was last time host connected
                # then, if the current time - last time >= limits close this process (probably the host is not hosting anymore)
                continue
            #lets check if is ok what the host has sent us
            if not self.client.is_body_ok_for_trainer_neat():
                exit(1)
            # first time receiving something
            # pay attention, if the host starts the same training closing its entire process
            # it will not have this identifier again, and this trainer will be constantly disconnected by the host
            # in that case this trainer should shut down the process and restart it
            if not self.client.identifier_is_set():
                self.client.get_body_identifier()
            # set the genomes structures
            self.client.get_genomes(self.sub_games)
            # set the innovation number of connections
            self.client.get_global_innovation_number_connections()
            # set the innovation number of nodes
            self.client.get_global_innovation_number_nodes()
            # set the number of genomes
            self.client.get_number_of_genomes()
            # retrieve the secondary sockets public keys
            secondary_sockets= self.client.get_link().decode('utf-8')
            identifier = self.client.get_py_identifier()
            # retrieve the environment name if we didn't yet
            if environment_name == None:
                environment_name = self.client.get_environment_name().decode('utf-8')
            # python can now see the number of genomes
            n_genomes = self.client.returns_n_genomes()
            # we use the socket already initialized
            if not self.first_iteration_distributed_training:
                self.set_file_descriptor(True)
            else:
                self.first_iteration_distributed_training = False
            # creating the environments:ù
            res = {'full':True}
            first_time = True
            n_genomes*=self.sub_games
            param_init = {'endpoint': self.environment_creation_endpoint, 'env_id':environment_name, 'n_instances':n_genomes, 'identifier':identifier}
            param_init_str = str(param_init)
            param_init_str = param_init_str.replace("'",'"')
            # maybe the client is already serving too  much clients
            # so we see if we can open other environments
            # if we can a res{'full':True} will be received
            # and we will wait for 1 sec to ask again
            #init the environments
            while 'full' in res and res['full']:
                if not first_time:
                    time.sleep(1)
                try:
                    res = self.send_http_request(param_init_str, secondary_sockets)
                    res = json.loads(res)
                except:
                    print("exiting")
                    exit(1)
            if 'ok' in res or res == None:
                last = datetime.now()
                self.set_file_descriptor(False)
                self.client.free_genomes()
                self.client.set_reset_body_when_reconnect()
                self.client.connect(self.remote_ip,self.remote_port)
                continue
            
            max_games = [0]*self.sub_games
            count = 0
            while count != self.max_number_of_games:
                for i in range(self.sub_games):
                    max_games[i]+=1
                    count+=1
                    if count == self.max_number_of_games:
                        break
            
            
            
            list_keys = list(res.keys())
            list_keys.sort()
            if n_genomes != len(list_keys):
                exit(1)
            
            #init some parameters
            list_states = []
            list_rewards = []
            steps = []
            games = []
            to_count_yet = []
            done = []
            game_done = []
            max_number_of_games = []
            
            for i in range(n_genomes):
                list_states.append(res[list_keys[i]])
                list_rewards.append(0)
                steps.append(0)
                games.append(0)
                to_count_yet.append(1)
                done.append(0)
                game_done.append(False)
                max_number_of_games.append(max_games[int(i/int(n_genomes/self.sub_games))])
            # training
            flag_break = False
            while True:
                list_output_to_keep = []
                for i in range(n_genomes):
                    if games[i] < max_number_of_games[i]-1 or games[i] < max_number_of_games[i] and steps[i] < self.max_number_of_steps-1 and not game_done[i]:
                        list_output_to_keep.append(1)
                    else:
                        list_output_to_keep.append(0)
                out = self.client.forward_genomes(list_states, self.threads, done, to_count_yet, list_keys, list_rewards, list_output_to_keep, self.sub_games)
                param_post = {}
                list_of_environments = []
                actions = []
                j = 0
                n_done = 0
                for i in range(n_genomes):
                    if steps[i] < self.max_number_of_steps and games[i] < max_number_of_games[i]:
                        steps[i]+=1
                        if game_done[i] or steps[i] == self.max_number_of_steps:
                            games[i]+=1
                            steps[i] = 0
                    if done[i] > 0:
                        to_count_yet[i] = 0
                    else:
                        if games[i] < max_number_of_games[i]:
                            list_of_environments.append(list_keys[i])
                            actions.append(out[j])
                            
                        else:
                            done[i] = 1
                            to_count_yet[i] = 0
                        j+=1
                length_env_to_count = len(list_of_environments)
                if length_env_to_count == 0:
                    break
                for i in range(length_env_to_count):
                    param_post[list_of_environments[i]] = actions[i]
                    param_post['endpoint'] = self.next_step_endpoint
                    param_post_str = str(param_post)
                    param_post_str = param_post_str.replace("'",'"')
                try:
                    res = self.send_http_request(param_post_str, secondary_sockets)
                    res2 = res
                    res = json.loads(res)
                except:
                    exit(1)
                if 'ok' in res or res == None:
                    last = datetime.now()
                    self.set_file_descriptor(False)
                    flag_break = True
                    self.client.free_genomes()
                    self.client.set_reset_body_when_reconnect()
                    self.client.connect(self.remote_ip,self.remote_port)
                    break
                list_got = list(res.keys())
                for i in range(n_genomes):
                    if list_keys[i] in list_got:
                        list_states[i] = res[list_keys[i]][0]
                        game_done[i] = res[list_keys[i]][1]
            if flag_break:
                continue
            self.client.set_values_back_in_body()
            last = datetime.now()
            self.set_file_descriptor(False)
